#!/usr/bin/env python

from functools import wraps

import hashlib
import logging
from optparse import OptionParser
from optparse import OptionValueError
import os
import re
import requests
import sys
import time
import urlparse

import libtorrent
import transmissionrpc

APPLICATION = 'seedclient/0.17'

logger = logging.getLogger(__name__)


def retry(ExceptionToCheck, tries=5, delay=3, backoff=2, logger=logger):
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck, e:
                    msg = "%s, Retrying in %d seconds..." % (str(e), mdelay)
                    if logger:
                        logger.warning(msg)
                    else:
                        print msg
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)
        return f_retry  # true decorator
    return deco_retry


class SeedClient:
    def __init__(self):
        self.logger = logger
        self.file_checksum = {}
        self.parser = OptionParser(
            usage="\n\n"
                  "START DOWNLOADING WITH MAGNET LINK\n"
                  "    %prog -d -m MAGNET [-v]\n\n"
                  "START SEEDING LOCAL FILE\n"
                  "    %prog -u -f FILENAME --tracker-url=TRACKER "
                  "[OPTIONS]\n\n"
                  "UPLOAD LOCAL FILE TO REMOTE HOST AND START SEEDING THERE\n"
                  "    %prog -p -f FILENAME --tracker-url=TRACKER "
                  "--seed-host=SEED_HOST[,SEED_HOST2[,SEED_HOST3]]\n\n"
        )

        self.parser.add_option(
            '-u', '--upload', action='store_true', dest='upload',
            help='upload file, start seeding and detach')

        self.parser.add_option(
            '-f', '--file', dest='file',
            help='Path to file to operate with'
        )

        self.parser.add_option(
            '-d', '--download', action='store_true', dest='download',
            help='download file, proceed seeding and detach with printed '
                 'localpath'
        )

        self.parser.add_option(
            '-m', '--magnet-link', dest='magnet',
            default=os.environ.get('MAGNET_LINK', None),
            action="callback", type="string", nargs=1,
            callback=self._parser_magnet_verify,
            help='download file with magnet link'
        )

        self.parser.add_option(
            '-o', '--output-dir', dest='output_dir',
            default=os.environ.get('SEEDCLIENT_OUTPUT_DIR', '/tmp'),
            help='Output directory for downloaded torrent'
        )

        self.parser.add_option(
            '-v', '--verbose', action='store_true', dest='verbose',
            default=os.environ.get('SEEDCLIENT_VERBOSE', False),
            help='print verbose information during work'
        )

        self.parser.add_option(
            '-p', '--publish', action='store_true', dest='publish',
            default=False, help='Publish torrent to seed-web'
        )

        self.parser.add_option(
            '--seed-host', dest='seed_host',
            default=os.environ.get('SEED_HOST', ''),
            help='Seed host'
        )

        self.parser.add_option(
            '--transmission-host', dest='transmission_host',
            default=os.environ.get('TRANSMISSION_HOST', '127.0.0.1'),
            help='Transmission RPC host'
        )

        self.parser.add_option(
            '--transmission-port', dest='transmission_port',
            default=os.environ.get('TRANSMISSION_PORT', '9091'),
            help='Transmission RPC port'
        )

        self.parser.add_option(
            '--transmission-user', dest='transmission_user',
            default=os.environ.get('TRANSMISSION_USER', 'transmission'),
            help='Transmission RPC user'
        )

        self.parser.add_option(
            '--transmission-password', dest='transmission_password',
            default=os.environ.get('TRANSMISSION_PASSWORD', 'transmission'),
            help='Transmission RPC password'
        )

        self.parser.add_option(
            '--tracker-url', dest='tracker_url',
            default=os.environ.get('TRACKER_URL', None),
            help='Torrent tracker url'
        )

        self.parser.add_option(
            '--http-root', dest='http_root',
            default=os.environ.get('HTTP_ROOT', None),
            help='Add webseed from this directory'
        )

        self.parser.add_option(
            '--force-set-symlink', action='store_true',
            dest='force_set_symlink', default=False,
            help='Deprecated: Remove file if it exists before setting symlink '
                 'after download is complete'
        )

        self.parser.add_option(
            '--commit', dest='commit',
            default=os.environ.get('COMMIT', None),
            help='Amount of successfull upload meant to treat seedclient run '
                 'successful'
        )

        self.parser.add_option(
            '--share-dir', dest='share_dir',
            default='/fuelweb-iso',
            help='Remote directory to push content, default: /fuelweb-iso'
        )

        (self.options, self.args) = self.parser.parse_args()

        handler = logging.StreamHandler(sys.stderr)
        if self.options.verbose:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.WARNING)

        formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        if self.options.upload or self.options.download:
            self._connect()

        if self.options.file:
            self.options.file = os.path.abspath(self.options.file)

    @staticmethod
    def _parser_magnet_verify(option, opt, value, parser):
        if re.match('^magnet:', value) or re.match(
            '^http(s?):\/\/.*\.torrent$', value
        ):
            setattr(parser.values, option.dest, value)
        else:
            raise OptionValueError(
                "'{0}' is not valid magnet link".format(value)
            )

    @retry(Exception, tries=5, delay=1, backoff=2)
    def _connect(self):
        try:
            self.logger.info('Connecting to Transmission RPC at %s:%s' % (
                self.options.transmission_host,
                self.options.transmission_port))
            self.tc = transmissionrpc.Client(
                address=self.options.transmission_host,
                port=self.options.transmission_port,
                user=self.options.transmission_user,
                password=self.options.transmission_password,
            )
            self.logger.info('Connection to Transmission RPC successfully '
                             'establised')
            return True
        except Exception as e:
            self.logger.warn('Connection failed: %s' % e.message)
            raise

    def _files(self, torrent):
        files = torrent.files()

        for id in files:
            output_dir = os.path.normpath(self.options.output_dir)
            download_dir = os.path.normpath(torrent.downloadDir)

            file = os.path.join(download_dir, files[id]['name'])
            if not os.path.exists(file):
                self.logger.debug('Source \'%s\' file do not exists!' % file)
                sys.exit(1)
            destfile = file
            self.logger.debug('File: %s ; Download directory: %s' % (
                file, torrent.downloadDir))
            if output_dir != download_dir:
                destfile = os.path.join(
                    output_dir,
                    files[id]['name']
                )

                if os.path.islink(destfile):
                    lfile = os.readlink(destfile)
                    if os.path.exists(lfile) and file == lfile:
                        self.logger.info('Skipping already existing \'%s\' '
                                         'symlink' % files[id]['name'],
                                         )
                    else:
                        self.logger.info('Removing broken \'%s\' symlink'
                                         % files[id]['name'],
                                         )
                        os.unlink(destfile)

                if os.path.exists(destfile) and not os.path.islink(destfile):
                    self.logger.info('The \'%s\' exists and is not a symlink, '
                                     'removing it' % files[id]['name'],
                                     )
                    os.unlink(destfile)

                if not os.path.exists(destfile):
                    self.logger.info('Settings symlink for file %s to %s' % (
                        file, output_dir))
                    os.symlink(file, destfile)
            print destfile
        exit(0)

    def _generate_checksum(self):
        self.logger.info('Generating checksums')
        for algo in ('md5',):  # 'sha1', 'sha256', 'sha512',):
            h = hashlib.new(algo)

            c = open('%s.%s' % (self.options.file, algo), 'wb')

            h.update(open(self.options.file, 'r').read())
            self.file_checksum[algo] = h.hexdigest()

            self.logger.info('%s checksum generated: %s' % (
                algo, self.file_checksum[algo]
            ))

            c.write('%s  %s' % (
                self.file_checksum[algo],
                os.path.basename(self.options.file)))
            c.close()

    def _generate_metadata(self):
        fs = libtorrent.file_storage()
        libtorrent.add_files(fs, self.options.file)
        fs.num_files()

        t = libtorrent.create_torrent(fs, 256 * 1024)

        if self.options.tracker_url:
            tier = 0
            for tracker in self.options.tracker_url.split(','):
                self.logger.info('Adding tracker %s ; tier = %s' % (
                    tracker, tier))
                t.add_tracker(tracker, tier)
                tier += 1

        if self.options.http_root:
            for webseed in self.options.http_root.split(','):
                url = os.path.join(
                    webseed,
                    os.path.basename(self.options.file))
                self.logger.info('Adding webseed %s' % url)
                t.add_url_seed(url)

        self.logger.info('Generating piece hashes')
        libtorrent.set_piece_hashes(t, os.path.dirname(self.options.file))
        t.set_creator(APPLICATION)

        self.logger.info('Generating metadata')
        metadata = t.generate()
        info = libtorrent.torrent_info(metadata)
        infohash = str(info.info_hash()).lower()
        self.logger.info('Infohash: %s' % infohash)

        self.logger.info('Writing torrent file %s.torrent' % self.options.file)
        fp = open('%s.torrent' % self.options.file, 'w')
        fp.write(libtorrent.bencode(metadata))
        fp.close()

        return (infohash, '%s.torrent' % self.options.file)

    @retry(Exception, tries=10, delay=1, backoff=2)
    def _status(self, torrent):
        while True:
            try:
                t = self.tc.get_torrent(torrent_id=torrent.id)
            except KeyError:
                self.logger.debug('Torrent list dump: %s' %
                                  self.tc.get_torrents()
                                  )
                raise

            try:
                peers = t._fields['peersConnected'].value
            except KeyError:
                peers = 0

            self.logger.info('Torrent %s ; status %s ; progress %s ; peers %s'
                             % (t.name, t.status, t.progress, peers)
                             )
            if t.progress == 100:
                self.logger.info('Torrent %s download complete' % torrent.name)
                self._files(t)
            else:
                time.sleep(3)

    def download(self):
        try:
            download_as_torrent = True
            torrent = self.tc.add_torrent(self.options.magnet)
            self.logger.info('Starting downloading torrent %s' % (
                self.options.magnet))
        except (transmissionrpc.error.TransmissionError,
                transmissionrpc.error.HTTPHandlerError) as exc:
            if str(exc) == 'Query failed with result "duplicate torrent".':
                for torrent in self.tc.get_torrents():
                    if torrent.magnetLink == self.options.magnet:
                        if torrent.progress == 100:
                            self.logger.info('Torrent is already downloaded')
                            self._files(torrent)
                        else:
                            self.logger.info('Torrent already added and '
                                             'currently downloading')
                            self._status(torrent)
            else:
                self.logger.error(exc)
                self.logger.info('Torrent could not be downloaded. '
                                 'Trying to fallback into HTTP download.')

                download_as_torrent = False
                self.fall_back_to_http_download()

        if download_as_torrent:
            self._status(torrent)

    def upload(self):
        if not self.options.file:
            print '-f|--file FILENAME is required'
            exit(1)

        if not os.path.exists(self.options.file):
            print '%s no such file or directory' % self.options.file
            exit(1)

        self.logger.info('Collecting data about %s' % self.options.file)

        infohash, torrent_file = self._generate_metadata()
        self._generate_checksum()
        download_dir = os.path.dirname(self.options.file)

        try:
            torrent = self.tc.get_torrent(infohash)
            self.logger.info('Torrent %s(%s) already exists' % (
                infohash, self.options.file))
        except KeyError as e:
            if e.message == 'Torrent not found in result':
                torrent = self.tc.add_torrent(
                    '%s.torrent' % self.options.file,
                    download_dir=download_dir)
                self.logger.info('Adding torrent %s(%s) to transmission' % (
                    infohash, self.options.file))
                self.logger.debug('Running verify for torrent %s' % infohash)
                self.tc.verify_torrent(torrent.id)
                self.logger.debug('Starting torrent %s' % infohash)
                self.tc.start_torrent(torrent.id)

                torrent = self.tc.get_torrent(torrent.id)
            else:
                raise

        # Missed webseeds from magnet link
        # https://code.google.com/p/libtorrent/issues/detail?id=599
        # magnet = str(libtorrent.make_magnet_uri(info)).lower()
        magnet = torrent.magnetLink

        self.logger.info('Magnet link: %s' % magnet)
        print(magnet)

        if self.options.publish:
            self.publish(self.options.file, torrent.magnetLink)

    def publish(self):
        if not self.options.file:
            print '-f|--file FILENAME is required'
            exit(1)

        if not os.path.exists(self.options.file):
            print '%s no such file or directory' % self.options.file
            exit(1)

        if not self.options.seed_host:
            print 'There is no seed hosts defined.'
            exit(1)

        self.logger.info('Starting publishing')
        self.logger.info('Generating torrent file')
        info_hash, torrent_file = self._generate_metadata()
        self._generate_checksum()

        seeds = self.options.seed_host.split(',')
        counter = 0
        for seed in seeds:
            self.logger.info('Starting upload %s to %s' % (
                self.options.file, seed))

            rfile = requests.put(
                'http://%s/%s/%s' % (
                    seed, self.options.share_dir,
                    os.path.basename(self.options.file)),
                data=open(self.options.file, 'rb'),
                headers={'User-Agent': APPLICATION},
            )

            if rfile.status_code == 201:
                counter += 1
                self.logger.info('File %s ; Status %s ; Seed host %s' % (
                    self.options.file, rfile.status_code, seed))
            else:
                self.logger.error(
                    'Publishing Failed. File %s ; Status %s ; Seed host %s' % (
                        self.options.file, rfile.status_code, seed))
                self.logger.debug('Content after failed request: %s' % (
                    rfile.content))

            rtorrent = requests.put(
                'http://%s/%s/%s' % (
                    seed,
                    self.options.share_dir,
                    os.path.basename(torrent_file)),
                data=open(torrent_file, 'rb'),
                headers={'User-Agent': APPLICATION},
            )

            if rtorrent.status_code == 201:
                counter += 1
                self.logger.info('File %s ; Status %s ; Seed host %s' % (
                    torrent_file, rtorrent.status_code, seed))
            else:
                self.logger.error(
                    'Publishing Failed. File %s ; Status %s ; Seed host %s' % (
                        self.options.file, rtorrent.status_code, seed))
                self.logger.debug('Content after failed request: %s' % (
                    rtorrent.content))

        if counter < len(seeds) * 2:
            self.logger.warning('Completed with errors.'
                                'Possibly this ain\'t fine')
        else:
            self.logger.info(
                'Completed without errors. %s uploaded to %s seeds' % (
                    self.options.file, len(seeds)))

        if self.options.commit:
            if counter < self.options.commit:
                sys.exit(1)

    def fall_back_to_http_download(self):
        '''Fallback to HTTP download if torrent could not be processed.'''
        magnet_prefix = 'magnet:?'
        if not self.options.magnet.startswith(magnet_prefix):
            self.logger.error("The link '%s' is not the magnet one.", self.options.magnet)
            sys.exit(1)

        output_dir = os.path.normpath(self.options.output_dir)
        web_seed_links = urlparse.parse_qs(
            self.options.magnet[len(magnet_prefix):]
        ).get('ws')

        if not web_seed_links:
            self.logger.error(
                "The 'ws' attribute is missed in the magnet link %s",
                self.options.magnet
            )
            sys.exit(1)

        for http_url in web_seed_links:
            file_name = re.search('.+/(.+)$', http_url).group(1)
            output_path = os.path.join(output_dir, file_name.rstrip('/'))

            self.logger.info('Trying to reach the %s seed link.', http_url)
            try:
                response = requests.get(http_url, stream=True)
            except requests.exceptions.ConnectionError as exc:
                self.logger.error(
                    'Seed link %s is unreachable: %s',
                    str(http_url),
                    str(exc)
                )
                continue

            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as exc:
                self.logger.error(
                    'Something gone wrong during HTTP download: %s',
                    str(exc)
                )
            else:
                with open(output_path, 'wb') as f:
                    for chunk in response.iter_content(chunk_size=1024):
                        if chunk:
                            f.write(chunk)

                self.logger.info('Download %s complete', output_path)
                sys.stdout.write(str(output_path))
                break


if __name__ == "__main__":
    sc = SeedClient()

    if sc.options.download:
        sc.download()
    elif sc.options.upload:
        sc.upload()
    elif sc.options.publish:
        sc.publish()
    else:
        sc.parser.print_help()
        exit(1)
